
module Phase5 where
import Prelude          as P
import Dictionnaire     as D
import Data.Map.Strict  as M
import Data.Set         as S
import Data.Maybe
import Control.Monad
import Data.Hashable
import Autorise


type Hash = Int
type Nom = Int


type TransitionsM mon = Dictionnaire Char (mon Nom)


data EtatM mon = EtatM { getFinal :: mon ()
                      , getTrans :: TransitionsM mon
                      }
data AutomateFiniM mon = AutomateFiniM { getEtats   :: Map Nom (EtatM mon)
                                      , getInitial :: mon Nom
                                      }

type AutomateFini = AutomateFiniM Maybe
type AutomateFiniND = AutomateFiniM []


-- Reconnaissance d'un mot par un automate
reconnise :: Monad mon => AutomateFiniM mon -> String -> mon ()
reconnise (AutomateFiniM etats nomM) s = do
     nom <- nomM
     reconniseE (etats ! nom) s  
     where
       reconniseE e ""          = getFinal e
       reconniseE e (c:w)       = do
       	      nom <- rechercherParCleM (getTrans e) c
              reconniseE (etats ! nom) w

-- créer un automate a partir d'une liste d'états avec leur booleen d'acceptation et d'une liste de transitions
-- l'état initial est le premier nom de la liste de noms
newAutomaton :: (Monad mon, Eq (mon Nom)) => [(Nom,mon ())] -> [(Nom,Char,mon Nom)] -> AutomateFiniM mon
newAutomaton noms tr = AutomateFiniM (creerEtats noms tr) $ return $ fst $ head noms

creerEtats :: (Monad mon, Eq (mon Nom)) => [(Nom,mon ())] -> [(Nom,Char,mon Nom)] -> Map Nom (EtatM mon)
creerEtats noms tr = mapWithKey (creerEtat tr) (M.fromList noms)

creerEtat :: (Monad mon, Eq (mon Nom)) => [(Nom,Char,mon Nom)] -> Nom -> mon () -> EtatM mon
creerEtat tr e b = EtatM b (insererListe creer transitions)
   where
     tr_e = P.filter (\(e',_,_)->e==e') tr
     transitions = fmap (\(_,c,n)->(c,n)) tr_e

-- type des automate infinis du TP2
data EtatD = EtatD Bool (Char->EtatD)
type AutomateD = EtatD

-- projeter sur le type du TP2
forgetFiniteness :: AutomateFini -> AutomateD
forgetFiniteness (AutomateFiniM etats Nothing)     = error "pas d'état initial"
forgetFiniteness (AutomateFiniM etats (Just init)) = forgetFinitenessEtat (etats ! init)
   where
   forgetFinitenessEtat (EtatM b tr) =
       EtatD (b/=Nothing) (enleverMaybe . (forgetFinitenessTransition tr))
   forgetFinitenessTransition tr c = do
       nom  <- rechercherParCleM tr c
       etat <- etats !? nom
       return (forgetFinitenessEtat etat)
   enleverMaybe (Just e) = e
   enleverMaybe Nothing  = EtatD False (\c-> enleverMaybe Nothing)


type NomND   = Int
type EtatND  = EtatM []
type Etat    = EtatM Maybe
type TransND = TransitionsM []

-- fonction de determinisation
determinise :: AutomateFiniND -> AutomateFini
determinise (AutomateFiniM etatsND initND) = AutomateFiniM etats init
   where
   init  = Just (nommer initND)
   etats = determiniseAux etatsND M.empty [initND] 

determiniseAux :: (Map NomND EtatND) -> Map Nom Etat -> [[NomND]] -> Map Nom Etat
determiniseAux etatsND acc [] = acc
determiniseAux etatsND acc (noms:l)
        | M.member nomDet acc = determiniseAux etatsND acc l
	| otherwise           = determiniseAux etatsND (M.insert nomDet etat acc) ((D.toListVal ft)++l)
	where
	  nomDet   = nommer noms
	  etat     = EtatM (estTerminalList es) (mapDico (Just . nommer) ft)
          es       = fmap (etatsND !) noms
	  ft       = fusionT es



estTerminalList :: [EtatND] -> Maybe ()
estTerminalList es = listToMaybe (getFinal =<< es)
 

regroupe :: [TransND] -> TransND
regroupe = (mapDico concat) . unionCons 

fusionT :: [EtatND] -> TransND
fusionT = regroupe . (fmap getTrans)

nommer :: [NomND] -> Nom
nommer = hash . S.toList . S.fromList



-- créer un automate qui reconnaît un caractère compris dans un segment donné
segment :: Char -> Char -> AutomateFiniND
segment c1 c2 = AutomateFiniM  (M.fromList [(hashWithSalt 5555 (c1,'a',c2),etat1) ,
                                            (hashWithSalt 5555(c1,'b',c2),etat2) ])
                               [hashWithSalt 5555 (c1,'a',c2)]
               where
	          etat1 = EtatM []  (D.insererIntervalle D.creer c1 c2 [hashWithSalt 5555(c1,'b',c2)])
	          etat2 = EtatM [()] D.creer


-- créer un automate qui reconnaît la somme des langages reconnus par les automates entres
(!+!) :: AutomateFiniND -> AutomateFiniND -> AutomateFiniND
e1 !+! e2 =
     AutomateFiniM (M.union etats1 etats2) (i1 ++ i2)
     where
      (AutomateFiniM etats1 i1) = modifNom e1 (\x -> hashWithSalt (x*235154) 'a')
      (AutomateFiniM etats2 i2) = modifNom e2 (\x -> hashWithSalt (x*235154) 'b')


modifNom :: (Monad mon, Eq (mon Nom)) => AutomateFiniM mon -> (Nom->Nom) -> AutomateFiniM mon
modifNom (AutomateFiniM etats i) f =
     AutomateFiniM (mapKeys f (fmap modifNomE etats)) (liftM f i)
     where
     modifNomE (EtatM b tr) = EtatM b (mapDico (liftM f) tr)


-- créer un automate qui reconnaît l'étoile du langage reconnu par l'automate entre
star :: AutomateFiniND -> AutomateFiniND
star e = AutomateFiniM (M.insert (-9999999999999) etat0 (fmap cont etats)) [(-9999999999999)]
     where
       cont (EtatM [] tr) = EtatM [] tr
       cont (EtatM l  tr) = EtatM l (regroupe [tr,transInit])
       AutomateFiniM etats i = modifNom e (\x -> hashWithSalt x 's')
       transInit = regroupe (fmap (getTrans . (etats !)) i)
       etat0 = EtatM [()] transInit


estTerminalListND :: [EtatND] -> [()]
estTerminalListND es = getFinal =<< es

-- créer un automate qui reconnaît la concaténation des langages reconnus par les automates entres
(!.!) :: AutomateFiniND -> AutomateFiniND -> AutomateFiniND
e1 !.! e2 =
     AutomateFiniM (M.union (fmap cont etats1) etats2) i1
     where
       cont (EtatM [] tr) = EtatM [] tr
       cont (EtatM l  tr) = EtatM (estTerminalListND  (fmap (etats2!) i2))
                                    (regroupe (tr:(fmap (getTrans . (etats2 !)) i2 )))
       AutomateFiniM etats1 i1 = modifNom e1 (\x -> hashWithSalt x 'a')
       AutomateFiniM etats2 i2 = modifNom e2 (\x -> hashWithSalt x 'b')


-- type des expressions régulières
data Regexp = Segment Char Char  |  Regexp :+: Regexp  |  Star Regexp  |  Regexp :.: Regexp |  Eps


-- construire un automate reconnaissant le même langage que l'expression régulière entrée
regexp2Automata :: Regexp -> AutomateFini
regexp2Automata = determinise . regexp2AutomataND

regexp2AutomataND :: Regexp -> AutomateFiniND
regexp2AutomataND (Segment c1 c2) = segment c1 c2
regexp2AutomataND (r1 :+: r2) = regexp2AutomataND r1 !+! regexp2AutomataND r2
regexp2AutomataND (r1 :.: r2) = regexp2AutomataND r1 !.! regexp2AutomataND r2
regexp2AutomataND (Star r) = star (regexp2AutomataND r)
regexp2AutomataND Eps =  AutomateFiniM (M.fromList [(999999999,EtatM [()] D.creer)]) [999999999]





-- Automates figés


type TransitionsFM mon = Dictionnaire Char (mon (EtatFM mon))
data EtatFM mon = EtatFM (mon ()) (TransitionsFM mon)
type AutomateFiniFM mon = mon (EtatFM mon)


figer :: (Monad mon) => AutomateFiniM mon -> AutomateFiniFM mon
figer (AutomateFiniM etats iM) = do
     i <- iM
     figerE (etats ! i)
  where
  figerE (EtatM b tr) = return $ EtatFM b (mapDicoUnsafe ((\n -> figerE (etats ! n)) =<<) tr)

reconiseF :: (Monad mon) => AutomateFiniFM mon -> String -> mon ()
reconiseF eM s = flip reconiseFAux s =<< eM
  where
  reconiseFAux (EtatFM b _) []     = b
  reconiseFAux (EtatFM _ tr) (c:w) = do
      etat <- rechercherParCleM tr c
      reconiseFAux etat w







-- minimiser un automate
minimisation :: AutomateFini -> AutomateFini
minimisation aut@(AutomateFiniM etats iM) = AutomateFiniM etatsMin (fmap (transfNoms !) iM)
  where
  transfNoms = transformation aut
  etatsMin = foldlWithKey (\acc n1 n2 -> M.insert n2 (transfEtat (etats ! n1)) acc) M.empty transfNoms
  transfEtat (EtatM b tr) = EtatM b (mapDico (fmap (transfNoms !)) tr)


transformation :: AutomateFini -> Map Nom Nom
transformation  (AutomateFiniM etats _) = minAux (size init) init
  where
  init = fmap (hash . getFinal) etats
  size = S.size . S.fromList . M.elems
  contr m (EtatM bM tr) = hash (bM, mapDico (m !) (D.mapMaybe id tr))
  minAux s acc
      | s == s'   = acc
      | otherwise = minAux s' res
      where
        res = fmap (contr acc) etats  
        s'  = size res



-- bonus Ph4

type BigString = [String]

executAll :: AutomateFini -> String -> Map Nom Nom
executAll (AutomateFiniM etats _) st = mapMaybeWithKey (\n _ -> executOne n st) etats
   where
   executOne :: Nom -> String -> Maybe Nom
   executOne n ""    = Just n
   executOne n (c:w) = do
       e  <- etats !? n
       n' <- rechercherParCleM (getTrans e) c
       executOne n' w

paralleleExecution :: AutomateFini -> BigString -> [Map Nom Nom]
paralleleExecution aut = parMapS (executAll aut)

parReconnise :: AutomateFini -> BigString -> Maybe ()
parReconnise aut@(AutomateFiniM etats iM) bs = do
  i <- iM
  n <- reduce i $ paralleleExecution aut bs
  getFinal (etats ! n)
  where
  reduce :: Nom -> [Map Nom Nom] -> Maybe Nom
  reduce = foldM (\n d -> d !? n)



-- a utiliser si librairie indisponible et dans la version rendu

parMapS :: (a -> Map Nom Nom) -> [a] -> [Map Nom Nom]
parMapS = fmap


-- a utiliser si librairie disponible (pas dans la version rendue
--
-- evalDico :: Strategie (Map Nom Nom)
-- evalDico = evalTraversable return
--
-- parMapS :: (a -> Map Nom Nom) -> [a] -> [Map Nom Nom]
-- parMapS = parMap evalDico




----------------- Projet avancé -----------------------------


-- Désolé, je n'ai pas eu le temps d'essayer de le rendre plus lisible :s

class Monad mon => Determinisable mon where
  distrib     :: [mon a] -> mon [a] 
  proj        :: [mon a] -> [a] 

type TransitionsNDM mon = Dictionnaire Char [mon NomND]
data EtatNDM mon = EtatNDM { getFinalND :: mon ()
                           , getTransND :: TransitionsNDM mon
                           }
data AutomateFiniNDM mon = AutomateFiniNDM (Map Nom (EtatNDM mon)) [mon Nom]



-- fonction de determinisation
determinise2 :: (Determinisable mon, Eq (mon Nom)) =>
                AutomateFiniNDM mon -> AutomateFiniM mon
determinise2 (AutomateFiniNDM etatsNDM initNDM) = AutomateFiniM etats initM
   where
   initM  = nommer2 initNDM
   etats = determiniseAux2 etatsNDM M.empty [proj initNDM] 



determiniseAux2 :: (Determinisable mon, Eq (mon Nom)) =>
                   (Map NomND (EtatNDM mon)) -> Map Nom (EtatM mon) -> [[NomND]] -> Map Nom (EtatM mon)
determiniseAux2 etatsND acc [] = acc
determiniseAux2 etatsND acc (noms:l)
        | M.member nomDet acc = determiniseAux2 etatsND acc l
	| otherwise           = determiniseAux2 etatsND (M.insert nomDet etat acc) ((fmap proj $ D.toListVal ft)++l)
	where
	  nomDet   = nommer noms
	  etat     = EtatM (estTerminalList2 es) (mapDico nommer2 ft)
          es       = fmap (etatsND !) noms
	  ft       = fusionT2 es



estTerminalList2 :: (Determinisable mon) => [EtatNDM mon] -> mon ()
estTerminalList2 es = listToMon =<< distrib (fmap getFinalND es)

listToMon :: (Monad mon) => [a] -> mon a
listToMon []    = fail ""
listToMon (x:l) = return x

regroupe2 :: (Eq (mon Nom)) => [TransitionsNDM mon] -> TransitionsNDM mon
regroupe2 = (mapDico concat) . unionCons 

fusionT2 :: (Eq (mon Nom)) => [EtatNDM mon] -> TransitionsNDM mon
fusionT2 = regroupe2 . (fmap getTransND)

nommer2 :: (Determinisable mon) => [mon NomND] -> mon Nom
nommer2 l = liftM (hash . S.toList . S.fromList) $ distrib l






-- créer un automate qui reconnaît la somme des langages reconnus par les automates entres
(<+>) :: (Determinisable mon, Eq (mon Nom)) =>
         AutomateFiniNDM mon -> AutomateFiniNDM mon -> AutomateFiniNDM mon
e1 <+> e2 =
     AutomateFiniNDM (M.union etats1 etats2) (i1 ++ i2)
     where
      (AutomateFiniNDM etats1 i1) = modifNom2 e1 (\x -> hashWithSalt (x*235154) 'a')
      (AutomateFiniNDM etats2 i2) = modifNom2 e2 (\x -> hashWithSalt (x*235154) 'b')

modifNom2 :: (Determinisable mon, Eq (mon Nom)) =>
             AutomateFiniNDM mon -> (Nom->Nom) -> AutomateFiniNDM mon
modifNom2 (AutomateFiniNDM etats i) f
       = AutomateFiniNDM (mapKeys f (fmap (modifNom2E f) etats)) (fmap (liftM f) i)
modifNom2E :: (Determinisable mon, Eq (mon Nom)) =>
              (Nom->Nom) -> EtatNDM mon -> EtatNDM mon
modifNom2E f (EtatNDM b tr) = EtatNDM b (mapDico (fmap (liftM f)) tr)

effet :: (Determinisable mon, Eq (mon Nom)) =>
         AutomateFiniND -> mon () -> AutomateFiniNDM mon
effet  (AutomateFiniM etats i) after = AutomateFiniNDM (fmap alterEtat etats) (fmap return i)
  where
  alterEtat (EtatM [] tr) = EtatNDM (return ()) (mapDico (fmap return) tr)
  alterEtat (EtatM _  tr) = EtatNDM after (mapDico (fmap return) tr)


-- créer un automate qui reconnaît l'étoile du langage reconnu par l'automate entre
starM :: (Determinisable mon, Eq (mon Nom)) =>
         AutomateFiniNDM mon -> AutomateFiniNDM mon 
starM e = undefined

-- AutomateFiniNDM (M.insert (-9999999999999) etat0 (fmap cont etats)) [return (-9999999999999)]
--      where
--        cont (EtatNDM eff tr) = EtatNDM eff $ regroupe2 [tr, mapDico (fmap (eff >>)) transInit]
--                              | otherwise = EtatNDM eff tr
--        AutomateFiniNDM etats i = modifNom2 e (\x -> hashWithSalt x 's')
--        transInit = liftM (\i' -> regroupe2 (fmap (getTransND . (etats !)) i)) (distrib i)
 --       etat0 = EtatNDM [return ()] transInit

(<.>) :: (Determinisable mon, Eq (mon Nom)) =>
         AutomateFiniNDM mon -> AutomateFiniNDM mon -> AutomateFiniNDM mon 
(<.>) = undefined

